/*  

    YABASIC ---  a simple Basic Interpreter
    written by Marc-Oliver Ihm 1995-99
    homepage: www.yabasic.de
    
    io.c --- code for screen and file i/o
    
    This file is part of yabasic and  may be copied only 
    under the terms of either the Artistic License or 
    the GNU General Public License (GPL), both of which 
    can be found at www.yabasic.de

*/


/* ------------- defines ---------------- */

#ifdef WINDOWS
#define WM_WANTKEY WM_APP
#endif


/* ------------- includes ---------------- */

#ifndef YABASIC_INCLUDED
#include "yabasic.h"       /* all prototypes and structures */
#endif


/* ------------- local defines ----------------- */

#define YC_BLACK 0
#define YC_WHITE 1
#define YC_RED 2
#define YC_BLUE 3
#define YC_GREEN 4
#define YC_YELLOW 5
#define YC_CYAN 6
#define YC_MAGENTA 7

#ifndef COLOR_BLACK
#define COLOR_BLACK	0
#define COLOR_RED	1
#define COLOR_GREEN	2
#define COLOR_YELLOW	3
#define COLOR_BLUE	4
#define COLOR_MAGENTA	5
#define COLOR_CYAN	6
#define COLOR_WHITE	7
#endif
#ifndef A_COLOR
#define A_COLOR 0xff00
#endif

/* ------------- external references ---------------- */

extern int yylineno;   /* current line number */
extern int yyparse();  /* call bison parser */


/* ------------- local functions ---------------- */

static int onechar(void); /* read one char from currentinstream */
static void backchar(int); /* put char back into stream */
static void readline(void); /* read one line from current stream */
static void curinit(void); /* initialize curses */
static void initcol(void); /* initialize curses colors */
#ifdef WINDOWS
static DWORD keythread(LPWORD); /* wait for key input from console */
static int is_valid_key(INPUT_RECORD *); /* check if input rec is valid key */
#endif
int name2yc(char *); /* convert a color name to an integer */
int yc2oc(int,int); /* convert a yabasic color to operating system color */
char *yc2short(int); /* convert yabasic colours to short colour name */
#ifdef UNIX
int oc2yc(int); /* convert an operating system color to yabasic color */
#endif

/* ------------- global variables ---------------- */

static int prompted; /* TRUE, if prompt is fresh */
FILE *streams[FOPEN_MAX]; /* file streams */
static FILE *cinstr; /* current stream for input */
static FILE *coutstr; /* current stream for output */
static char linebuffer[INBUFFLEN]; /* buffer for one line of input */
int curinized=FALSE; /* true, if curses has been initialized */
static char *currchar; /* current char to read */
static short stdfc; /* standard foreground color of window */
static short stdbc; /* standard background color of window */
#ifdef UNIX
static int winpid=-1;
static int termpid=-1;
#else
HANDLE wantkey=INVALID_HANDLE_VALUE; /* mutex to signal key desire */
HANDLE gotkey=INVALID_HANDLE_VALUE; /* mutex to signal key reception */
HANDLE wthandle=INVALID_HANDLE_VALUE; /* handle of win thread */
HANDLE kthandle=INVALID_HANDLE_VALUE; /* handle of inkey thread */
DWORD ktid; /* id of inkey thread */
int LINES=0; /* number of lines on screen */
int COLS=0; /* number of columns on screen */
HANDLE ConsoleInput; /* handle for console input */
HANDLE ConsoleOutput; /* handle for console output */
#endif


/* ------------- functions ---------------- */

void create_print(char type) /* create command 'print' */
{
  struct command *cmd;
  
  cmd=add_command(cPRINT,NULL);
  cmd->pointer=my_malloc(sizeof(int));
  /* store type of print  */
  cmd->tag=type;
}


void print(struct command *cmd) /* print on screen */
{
  int type;
  struct stackentry *p,*q;
  static int last='n';
  char *s;
  int x,y;
#ifdef WINDOWS
  CONSOLE_SCREEN_BUFFER_INFO csbi;
  SMALL_RECT rect;
  SMALL_RECT clip;
  COORD to;
  CHAR_INFO ci;
#endif
  
  type=cmd->tag;
  switch(type) {
  case 'n':  /* print newline */
    if (curinized && coutstr==stdout) {
#ifdef WINDOWS
      GetConsoleScreenBufferInfo(ConsoleOutput,&csbi);
      x=csbi.dwCursorPosition.X;
      y=csbi.dwCursorPosition.Y;
      if (y>=LINES-1) {
        rect.Left=0;
        rect.Right=COLS-1;
        rect.Top=1;
        rect.Bottom=LINES-1;
        clip.Left=0;
        clip.Right=COLS-1;
        clip.Top=0;
        clip.Bottom=COLS-1;
        to.X=0;
        to.Y=0;
        ci.Char.AsciiChar=' ';
        ci.Attributes=0;
        ScrollConsoleScreenBuffer(ConsoleOutput,&rect,&clip,to,&ci);
        to.X=0;
        to.Y=LINES-1;
        SetConsoleCursorPosition(ConsoleOutput,to);
      } else {
        to.X=0;
        to.Y=y+1;
        SetConsoleCursorPosition(ConsoleOutput,to);
      }
      break;
#else
      getyx(stdscr,y,x);
      if (y>=LINES-1) {
        scrl(1);
        y=y-1;
      }
      move(y+1,0);
      refresh();
      break;
#endif
    } else {
      string[0]='\n';
      string[1]='\0';
    }
    onestring(string);
    break;
  case 't':  /* print tab */
    string[0]='\t';
    string[1]='\0';
    onestring(string);
    break;
  case 'd':  /* print double value */
    p=pop(stNUMBER);
    sprintf(string,"%s%g",(last=='d')?" ":"",p->value);
    onestring(string);
    break;
  case 'u':
    p=pop(stSTRING);
    q=pop(stNUMBER);
    type='d';
    s=string;
    if (last=='d') {
      *s=' ';
      s++;
    }
    if (!myformat(s,q->value,p->pointer)) {
      sprintf(string,"'%s' is not a valid format",(char *)p->pointer);
      error(ERROR,string);
      break;
    }
    onestring(string);
    break;
  case 's': 
    p=pop(stSTRING);
    onestring((char *)p->pointer);
    break;
  }
  last=type;
}


void mymove() /* move to specific position on screen */
{
  int x,y;
#ifdef WINDOWS
  COORD coord;
#endif
  
  y=(int)pop(stNUMBER)->value;
  if (y<0) y=0;
  if (y>LINES-1) y=LINES-1;
  x=(int)pop(stNUMBER)->value;
  if (x<0) x=0;
  if (x>COLS-1) x=COLS-1;
  if (!curinized) {
    error(ERROR,"need to call 'clear screen' first");
    return;
  }
#ifdef UNIX
  move(y,x);
  refresh();
#else
  coord.X=x;
  coord.Y=y;
  SetConsoleCursorPosition(ConsoleOutput,coord);
#endif
}


void clearscreen() /* clear entire screen */
{
#ifdef WINDOWS
  DWORD written; /* number of chars actually written */
  COORD coord; /* coordinates to start writing */
#endif
  
  if (!curinized) curinit();
#ifdef UNIX
  clear();
  refresh();
#else
  coord.X=0;
  coord.Y=0;
  FillConsoleOutputCharacter(ConsoleOutput,' ',LINES*COLS,
    coord,&written);
  FillConsoleOutputAttribute(ConsoleOutput,
    FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE,LINES*COLS,
    coord,&written);
  SetConsoleCursorPosition(ConsoleOutput,coord);
#endif
}


static void curinit(void) /* initialize curses */
{
#ifdef WINDOWS
  CONSOLE_SCREEN_BUFFER_INFO coninfo; /* receives console size */
#endif
  
#ifdef UNIX
  initscr();
  initcol();
  setscrreg(0,LINES);
  scrollok(stdscr,TRUE);
  leaveok(stdscr,TRUE);
  keypad(stdscr,TRUE);
  intrflush(stdscr,FALSE);
  curs_set(0);
  def_prog_mode();
#else
  GetConsoleScreenBufferInfo(ConsoleOutput,&coninfo);
  COLS=coninfo.dwSize.X;
  LINES=coninfo.dwSize.Y;
#endif
  curinized=TRUE;
}


char *inkey(double maxtime) /* get char from keyboard */
{
  char *skey;
#ifdef WINDOWS
  int ms; /* number of milliseconds to wait*/
  DWORD oflags; /* saves normal state of console input buffer */
  DWORD flags; /* new input mode for console input buffer */
  static char conkeybuff[100],winkeybuff[100];
#else
  fd_set readfds;
  struct timeval tv;
  int maxfd;
  int winfd[2],termfd[2];
  char retkey[100];
  int status,ret;
#endif
  
  
  if (maxtime>=0.0 && maxtime<0.01) maxtime=0.01;
  if (!curinized) {
#ifdef UNIX
    error(ERROR,"need to call 'clear screen' first");
    goto sane_state;
#endif
  }
#ifdef UNIX
  retkey[0]='\0';
  winfd[0]=winfd[1]=termfd[0]=termfd[1]=-1;
  winpid=termpid=-1;
  if (pipe(winfd) || pipe(termfd)) {
    error(ERROR,"Couldn't open pipes");
    goto sane_state;
  }
    
    winpid=fork();
    if (winpid==0) {
      /* this is the child */
      signal(SIGINT,SIG_DFL);
      signal(SIGFPE,SIG_DFL);
      signal(SIGSEGV,SIG_DFL);
      retkey[0]='\0';
      if (winopened) getwinkey(retkey);
      if (*retkey) write(winfd[1],retkey,strlen(retkey));
      wait(&status);
      exit(0);
    } else if (winpid==-1) {
      error(ERROR,"Couldn't fork child");
      goto sane_state;
    }

    termpid=fork();
    if (termpid==0) {
      /* this is the child */
      signal(SIGINT,SIG_DFL);
      signal(SIGFPE,SIG_DFL);
      signal(SIGSEGV,SIG_DFL);
      retkey[0]='\0';
      noecho();
      cbreak();
      timeout(-1);
      gettermkey(retkey);
      if (*retkey) write(termfd[1],retkey,strlen(retkey));
      wait(&status);
      exit(0);
    } else if (termpid==-1) {
      error(ERROR,"Couldn't fork child");
      goto sane_state;
    }
    
    FD_ZERO(&readfds);
    FD_SET(termfd[0],&readfds);
    FD_SET(winfd[0],&readfds);
    fflush(stdout);
    maxfd=(termfd[0]>winfd[0])?termfd[0]:winfd[0];
    if (maxtime>=0) {
      tv.tv_sec=(int)maxtime;
      tv.tv_usec=(maxtime-(int)maxtime)*1000000.;
      ret=select(maxfd+1,&readfds,NULL,NULL,&tv);
    } else {
      ret=select(maxfd+1,&readfds,NULL,NULL,NULL);
    }
    if (ret==-1) {
      error(ERROR,"select failed");
      goto sane_state;
    } 
    
    if (FD_ISSET(termfd[0],&readfds)) {
      ret=read(termfd[0],retkey,100);
    } else if (FD_ISSET(winfd[0],&readfds)) {
      ret=read(winfd[0],retkey,100);
      if (!strncmp("MB",retkey,2)) getmousexyb(retkey,&mousex,&mousey,&mouseb);
    }
    else ret=0;
    retkey[ret]='\0';
sane_state:
    skey=retkey;
    if (winfd[0]>0) close(winfd[0]);
    if (winfd[1]>0) close(winfd[1]);
    if (termfd[0]>0) close(termfd[0]);
    if (termfd[1]>0) close(termfd[1]);
    if (termpid>0) {
      kill(termpid,SIGTERM);
      waitpid(termpid,&status,0);
    }
    if (winpid>0) {
      kill(winpid,SIGTERM);
      waitpid(winpid,&status,0);
    }
    reset_prog_mode(); /* prepare for input afterwards */
    
#elif WINDOWS
    /* create event to signal key */
    if (gotkey==INVALID_HANDLE_VALUE) gotkey=CreateEvent(NULL,FALSE,FALSE,NULL);
    if (wantkey==INVALID_HANDLE_VALUE) wantkey=CreateEvent(NULL,TRUE,FALSE,NULL);
    
    conkeybuff[0]=winkeybuff[0]='\0';
    
    GetConsoleMode(ConsoleInput,&oflags);
    flags=oflags&~(ENABLE_LINE_INPUT|ENABLE_ECHO_INPUT);
    SetConsoleMode(ConsoleInput,flags);
    
    /* create thread to observe console */
    if (kthandle==INVALID_HANDLE_VALUE) {
      kthandle=CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)keythread,
        (LPVOID)conkeybuff,0,(LPDWORD)&ktid);
    }
    /* signal, that we want a key */
    if (winopened) {
      SendMessage(window,WM_WANTKEY,0,(LPARAM)winkeybuff);
    }
    SetEvent(wantkey);
    ms=(int)(1000*maxtime);
    if (ms<0)
      WaitForSingleObject(gotkey,INFINITE);
    else 
      WaitForSingleObject(gotkey,ms);
    
    ResetEvent(wantkey);
    SetConsoleMode(ConsoleInput,oflags);
    if (*winkeybuff) 
      skey=winkeybuff;
    else
      skey=conkeybuff;
#endif
    return my_strdup(skey);
}


#ifdef WINDOWS
static DWORD keythread(LPWORD lparam) /* wait for key input from console */
{
  INPUT_RECORD inrec; /* for reading key-event */
  int key,skey;
  int num;
  char *keybuff;
  HANDLE evn[2];
  
  keybuff=(char *)lparam;
  evn[0]=wantkey;
  evn[1]=ConsoleInput;
  do {
    do {
      WaitForMultipleObjects(2,evn,TRUE,INFINITE);
      ReadConsoleInput(ConsoleInput,&inrec,1,&num);
    } while(!is_valid_key(&inrec));	
    ResetEvent(wantkey);
    if (isprint(inrec.Event.KeyEvent.uChar.AsciiChar)) {
      keybuff[0]=inrec.Event.KeyEvent.uChar.AsciiChar;
      keybuff[1]='\0';
    }
    else {
      key=inrec.Event.KeyEvent.wVirtualKeyCode;
      skey=-1;
      switch(key) {
      case 0x1b: skey=kESC;break;
      case 0x0d: skey=kENTER;break;
      case 0x09: skey=kTAB;break;
      case 0x21: skey=kSCRNUP;break;
      case 0x22: skey=kSCRNDOWN;break;
      case 0x70: skey=kF1;break;
      case 0x71: skey=kF2;break;
      case 0x72: skey=kF3;break;
      case 0x73: skey=kF4;break;
      case 0x74: skey=kF5;break;
      case 0x75: skey=kF6;break;
      case 0x76: skey=kF7;break;
      case 0x77: skey=kF8;break;
      case 0x78: skey=kF9;break;
      case 0x79: skey=kF10;break;
      case 0x7a: skey=kF11;break;
      case 0x7b: skey=kF12;break;
      case 0x24: skey=kHOME;break;
      case 0x23: skey=kEND;break;
      case 0x2d: skey=kINS;break;
      case 0x2e: skey=kDEL;break;
      case 0x08: skey=kBACKSPACE;break;
      case 0x27: skey=kRIGHT;break;
      case 0x25: skey=kLEFT;break;
      case 0x28: skey=kDOWN;break;
      case 0x26: skey=kUP;break;
      default:
        sprintf(keybuff,"key%x",key);
      }
      if (skey>0) strcpy(keybuff,ykey[skey]);
    }
    SetEvent(gotkey);
  }while(TRUE);
  
  return 0;
}


static int is_valid_key(INPUT_RECORD *rec) /* check if input rec contains valid key */
{
  if (rec->EventType!=KEY_EVENT || 
    !rec->Event.KeyEvent.bKeyDown ||
    rec->Event.KeyEvent.wVirtualKeyCode==VK_SHIFT ||
    rec->Event.KeyEvent.wVirtualKeyCode==VK_CONTROL) {
    return FALSE;
  }
  return TRUE;
}
#endif


char *replace(char *string) /* replace \n,\a, etc. */
{
  char *from,*to;
  
  from=to=string;
  while(*from) {
    if (*from=='\\') {
      from++;
      switch(*from) {
      case 'n':	*to='\n';break;
      case 't':	*to='\t';break;
      case 'v':	*to='\v';break;
      case 'b':	*to='\b';break;
      case 'r':	*to='\r';break;
      case 'f':	*to='\f';break;
      case 'a':	*to='\a';break;
      case '\\': *to='\\';break;
      case '\?': *to='\?';break;
      case '\'': *to='\'';break;
      case '\"': *to='\"';break;
      default:
        *to='\\';
        to++;
        *to=*from;
      }
    }
    else
      *to=*from;
    from++;
    to++;
  }
  *to='\0';
  return string;
}


void create_myopen(int num) /* create command 'myopen' */
{
  struct command *cmd;
  
  cmd=add_command(cOPEN,NULL);
  cmd->tag=num;
}


void myopen(struct command *cmd) /* open specified file for given name */
{
  FILE *handle;
  int stream,i;
  char *name;
  char *mode;
  char **pmode;
  static char *valid_modes[]={"r","w","a","rb","wb","ab",""};
  struct stackentry *p;
  
  
  if (cmd->tag==3 || cmd->tag==-2)
    mode=pop(stSTRING)->pointer;
  else 
    mode="a";
  name=my_strdup(pop(stSTRING)->pointer);
  if (cmd->tag>0) {
    stream=(int)pop(stNUMBER)->value;
  } else {
    stream=0;
    for(i=1;i<FOPEN_MAX-4;i++) {
      if (!streams[i]) {
	stream=i;
	break;
      }
    }
    if (!stream) {
      error(ERROR,"reached maximum number of open files");
      return;
    }
  }
  p=push();
  p->value=0.;
  p->type=stNUMBER;
  if (badstream(stream)) {
    errorcode=1;
    free(name);
    return;
  }
  if (streams[stream]!=NULL) {
    sprintf(errorstring,"stream already in use");
    errorcode=2;
    free(name);
    return;
  }
  for(pmode=valid_modes;**pmode;pmode++) if (!strcmp(*pmode,mode)) break;
  if (!**pmode) {
    sprintf(errorstring,"\'%s\' is not a valid filemode",mode);
    errorcode=3;
    free(name);
    return;
  }
  
  handle=fopen(name,mode);
  if (handle==NULL) {
    sprintf(errorstring,"could not open '%s'",name);
    errorcode=4;
    free(name);
    return;
  }
  streams[stream]=handle;
  errorcode=0;
  free(name);
  p->value=stream;
}


void checkopen(void) /* check, if open has been sucessfull */
{
  double result;
  
  result=pop(stNUMBER)->value;
  if (!result) {
    error(ERROR,errorstring);
  } 
}


void myclose(void) /* close the specified stream */
{
  int s;
  
  s=(int)pop(stNUMBER)->value;
  if (badstream(s)) return;
  if (streams[s]==NULL) {
    sprintf(string,"Stream %d already closed",s);
    error(WARNING,string);
    return;
  }
  fclose(streams[s]);
  streams[s]=NULL;
}


void create_myswitch(int stream) /* create command myswitch */
{
  struct command *cmd;
  
  cmd=add_command(cSWITCH,NULL);
  cmd->args=stream;
}


void myswitch(int stream) /* switch to specified stream */
{
  if (stream<0) stream=(int)pop(stNUMBER)->value;

  if (!stream) {
    cinstr=stdin;
    coutstr=stdout;
  } else {
    cinstr=streams[stream]; /* switch to stream */
    coutstr=cinstr;
    if (streams[stream]==NULL) {
      sprintf(string,"Stream %d not opened",stream);
      error(ERROR,string);
      return;
    } 
  }
  return;
}


void testeof(struct command *cmd) /* close the specified stream */
{
  int s,c;
  struct stackentry *result;
  
  s=(int)pop(stNUMBER)->value;
  if (badstream(s)) return;
  result=push();
  result->type=stNUMBER;
  if (streams[s]==NULL) {
    result->value=1.;
    return;
  }
  c=getc(streams[s]);
  if (c==EOF) {
    result->value=1.;
    return;
  }
  
  result->value=0.;
  ungetc(c,streams[s]);
  return;
}


int badstream(int stream) /* test for valid stream id */
{
  if (stream>FOPEN_MAX-4 || stream<1) {
    sprintf(string,"Can handle only streams from 1 to %d",FOPEN_MAX-4);
    error(ERROR,string);
    return TRUE;
  }
  return FALSE;
}


void create_myread(char type,int rl) /* create command 'read' */
{
  struct command *cmd;
  
  cmd=add_command(cREAD,NULL);
  cmd->args=rl; /* true, if read should go til eol */
  cmd->tag=type; /* can be 'd' or 's' */
}


void myread(struct command *cmd) /* read string or double */
{
  double d;
  char buffer[INBUFFLEN]; /* buffer with current input */
  int numread; /* number of bytes read */
  int tileol; /* true, if read should go til end of line */
  struct stackentry *s;
  int currch; /* current character */
  
  numread=0; /* no chars read'til now */
  buffer[0]='\0';
  tileol=cmd->args;
  
  /* skip leading whitespace */
  if (!tileol) {
    do {
      currch=onechar();
    } while(currch==' ' || currch=='\t');
    /* put back last char */
    if (currch!=EOF && currch!='\0') backchar(currch);
    if (currch=='\0' || currch==EOF) goto done;
  }      
  
  /* read chars */
  do {
    currch=onechar();
    buffer[numread]=currch;
    numread++;
  } while(((tileol && currch!='\0') ||
	   (!tileol && currch!=' ' &&
	    currch!='\t' && currch!='\0')) &&
      currch!=EOF && numread<INBUFFLEN);
  
  /* put back last char */
  if (currch!=EOF && currch!='\0') backchar(currch);
  /* and remove it from buff */
  if (currch!=EOF) numread--;
  buffer[numread]='\0';
  if (currch=='\0' || currch==EOF) goto done;
  
  /* skip trailing whitespace */
  if (!tileol) {
    do {
      currch=onechar();
    } while(currch==' ' || currch=='\t');
    if (currch!=EOF && currch!='\0') backchar(currch);
  }
  
done:
  if (cmd->tag=='s') { /* read string */
    s=push();
    s->type=stSTRING;
    s->pointer=my_strdup(buffer);}
  else { /* read double */
    s=push();
    s->type=stNUMBER;
    s->value=0.0;
    if (buffer[0] && (sscanf(buffer,"%lf",&d)==1)) s->value=d;
  }
}


static void readline(void) /* read one line from current stream */
{  
#ifdef UNIX
  char *nl; /* position of newline */
  int x,y;
#else
  int read;
#endif
  
  memset(linebuffer,'\0',INBUFFLEN);
#ifdef UNIX
  if (curinized && cinstr==stdin) {
    getyx(stdscr,y,x);
#ifdef HAVE_GETNSTR
    getnstr(linebuffer,INBUFFLEN);
#else
    getstr(linebuffer);
#endif
    if ((nl=strchr(linebuffer,'\0'))) *nl='\n';
    if (y>=LINES-1) scrl(1);
    refresh();
  }
#else
  if (curinized && cinstr==stdin) {
    FlushConsoleInputBuffer(ConsoleInput);
    ReadConsole(ConsoleInput,linebuffer,INBUFFLEN,&read,NULL);
    if (read>=2) {
      linebuffer[read-2]='\n';
      linebuffer[read-1]='\0';
    }
  }
#endif
  else {
    fgets(linebuffer,INBUFFLEN,cinstr);
  }
  currchar=linebuffer;
  prompted=FALSE;
}


static int onechar() /* read one char from cinstr */
{
  int ch;
  
  if (cinstr==stdin) {
    if (!currchar || !*currchar) {
      readline();
    }
    currchar++;
    ch=*(currchar-1);
  }
  else
    ch=fgetc(cinstr);
  
  if (ch=='\n' || ch==EOF)
    return '\0';
  else
    return ch;
}


static void backchar(int ch) /* put char back into stream */
{
  if (cinstr==stdin) {
    if (currchar>linebuffer) currchar--;
  }
  else {
    ungetc(ch,cinstr);
  }
}


void chkprompt() /* print an intermediate prompt if necessary */
{
  if (cinstr==stdin && (!currchar || !*currchar) && !prompted) onestring("?");
}


void create_onestring(char *str) /* create command 'onestring' */
{
  struct command *cmd;
  
  cmd=add_command(cONESTRING,NULL);
  cmd->pointer=my_strdup(str);
}


void onestring(char *string) /* write string to file */
{
#ifdef UNIX
  if (curinized && coutstr==stdout) {
    addstr(string);
    refresh();
  }
#else
  DWORD len,written;
  
  if (curinized && coutstr==stdout) {
    len=strlen(string);
    WriteConsole(ConsoleOutput,string,len,&written,NULL);
  }	
#endif
  else
  {
    fprintf(coutstr,"%s",string);
    fflush(coutstr);
  }
  prompted=TRUE;
}


void create_colour(int flag) /* create command 'colour' */
{
  struct command *c;
  
  c=add_command(cCOLOUR,NULL);
  c->args=flag;
}


void colour(struct command *cmd) /* switch on colour */
{
  char *fore=NULL,*back=NULL,*p;
  int fc,bc;

  if (cmd->args && !curinized) {
    error(ERROR,"need to call 'clear screen' first");
    return;
  }
  if (cmd->args==0) {
    if (!curinized) return;
#ifdef UNIX
    if (has_colors()) 
      attrset(A_NORMAL|COLOR_PAIR(stdfc*8+stdbc));
    else
      attrset(A_NORMAL);
    return;
#else
    SetConsoleTextAttribute(ConsoleOutput,FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_BLUE);
    return;
#endif
  } else if (cmd->args==1) {
#ifdef UNIX
    if (has_colors()) 
      attrset(A_NORMAL|COLOR_PAIR(stdbc*8+stdfc));
    else {
      attrset(A_REVERSE);
      return;
    }
#else
    SetConsoleTextAttribute(ConsoleOutput,BACKGROUND_RED|BACKGROUND_GREEN|BACKGROUND_BLUE);
    return;
#endif
  } else { /* decode colours */
#ifdef UNIX
    if (!has_colors()) {
      pop(stSTRING);
      if (cmd->args==3) pop(stSTRING);
      attrset(A_REVERSE);
      return;
    }
#endif
    if (cmd->args==2) {
      back=NULL;
      fore=pop(stSTRING)->pointer;
      for(p=fore;*p;p++) *p=tolower(*p);
    } else {
      back=pop(stSTRING)->pointer;
      for(p=back;*p;p++) *p=tolower(*p);
      fore=pop(stSTRING)->pointer;
      for(p=fore;*p;p++) *p=tolower(*p);
    }
    fc=name2yc(fore);
    if (fc<0) {
      sprintf(string,"unknown colour: '%s'",fore);
      error(ERROR,string);
    }
    bc=stdbc;
    if (back) {
      bc=name2yc(back);
      if (fc<0) {
	sprintf(string,"unknown colour: '%s'",back);
	error(ERROR,string);
      }
    }
#ifdef UNIX
    attrset(COLOR_PAIR(fc*8+bc));
#else
    SetConsoleTextAttribute(ConsoleOutput,(WORD)(yc2oc(fc,TRUE)|yc2oc(bc,FALSE)));
#endif
  }
}


static void initcol(void) /* initialize curses colors */
{
  static int first=TRUE;
#ifdef UNIX
  int i,j,col;
  short f,b;
#else
  CONSOLE_SCREEN_BUFFER_INFO csbi;
#endif

  if (!first) return;
  first=FALSE;
#ifdef UNIX
  if (!has_colors()) return;
  start_color();
  
  for(i=0;i<8;i++) {
    for(j=0;j<8;j++) {
      if (!i && !j) continue;
      init_pair(i*8+j,yc2oc(i,TRUE),yc2oc(j,FALSE));
    }
  }
  col=inch()&A_COLOR;
  pair_content(col,&f,&b);
  stdfc=oc2yc(f);
  stdbc=oc2yc(b);
  bkgdset(COLOR_PAIR(stdfc*8+stdbc));
#else
  GetConsoleScreenBufferInfo(ConsoleOutput,&csbi);
  stdfc=csbi.wAttributes & (FOREGROUND_BLUE|FOREGROUND_GREEN|FOREGROUND_RED);
  stdbc=csbi.wAttributes & (BACKGROUND_BLUE|BACKGROUND_GREEN|BACKGROUND_RED);
#endif
}


int name2yc(char *name) /* convert a color name to an integer */
{
  char *c;

  for(c=name;*c;c++) *c=tolower(*c);
  if (!strcmp(name,"black") || !strcmp(name,"bla")) return YC_BLACK;
  if (!strcmp(name,"white") || !strcmp(name,"whi")) return YC_WHITE;
  if (!strcmp(name,"red") || !strcmp(name,"red")) return YC_RED;
  if (!strcmp(name,"blue") || !strcmp(name,"blu")) return YC_BLUE;
  if (!strcmp(name,"green") || !strcmp(name,"gre")) return YC_GREEN;
  if (!strcmp(name,"yellow") || !strcmp(name,"yel")) return YC_YELLOW;
  if (!strcmp(name,"cyan") || !strcmp(name,"cya")) return YC_CYAN;
  if (!strcmp(name,"magenta") || !strcmp(name,"mag")) return YC_MAGENTA;
  return -1;
}


int yc2oc(int yc,int fore) /* convert a yabasic color to operating system color */
{
#ifdef UNIX
  fore=0; /* stop gcc from complaining */
  if (yc==YC_BLACK) return COLOR_BLACK;
  if (yc==YC_WHITE) return COLOR_WHITE;
  if (yc==YC_RED) return COLOR_RED;
  if (yc==YC_BLUE) return COLOR_BLUE;
  if (yc==YC_GREEN) return COLOR_GREEN;
  if (yc==YC_YELLOW) return COLOR_YELLOW;
  if (yc==YC_CYAN) return COLOR_CYAN;
  if (yc==YC_MAGENTA) return COLOR_MAGENTA;
#else
  if (fore) { 
    if (yc==YC_BLACK) return 0;
    if (yc==YC_WHITE) return FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE;
    if (yc==YC_RED) return FOREGROUND_RED;
    if (yc==YC_BLUE) return FOREGROUND_BLUE;
    if (yc==YC_GREEN) return FOREGROUND_GREEN;
    if (yc==YC_YELLOW) return FOREGROUND_GREEN | FOREGROUND_RED;
    if (yc==YC_CYAN) return FOREGROUND_GREEN | FOREGROUND_BLUE;
    if (yc==YC_MAGENTA) return FOREGROUND_BLUE | FOREGROUND_RED;
  } else {
    if (yc==YC_BLACK) return 0;
    if (yc==YC_WHITE) return BACKGROUND_RED | BACKGROUND_GREEN | BACKGROUND_BLUE;
    if (yc==YC_RED) return BACKGROUND_RED;
    if (yc==YC_BLUE) return BACKGROUND_BLUE;
    if (yc==YC_GREEN) return BACKGROUND_GREEN;
    if (yc==YC_YELLOW) return BACKGROUND_GREEN | BACKGROUND_RED;
    if (yc==YC_CYAN) return BACKGROUND_GREEN | BACKGROUND_BLUE;
    if (yc==YC_MAGENTA) return BACKGROUND_BLUE | BACKGROUND_RED;
  }
#endif
  return -1; 
}


int oc2yc(int oc) /* convert an operating system color to yabasic color */
{
#ifdef UNIX
  if (oc==COLOR_BLACK) return YC_BLACK;
  if (oc==COLOR_WHITE) return YC_WHITE;
  if (oc==COLOR_RED) return YC_RED;
  if (oc==COLOR_BLUE) return YC_BLUE;
  if (oc==COLOR_GREEN) return YC_GREEN;
  if (oc==COLOR_YELLOW) return YC_YELLOW;
  if (oc==COLOR_CYAN) return YC_CYAN;
  if (oc==COLOR_MAGENTA) return YC_MAGENTA;
#else
  if (oc==0) return YC_BLACK;
  if (oc==(FOREGROUND_RED|FOREGROUND_BLUE|FOREGROUND_GREEN) || 
      oc==(BACKGROUND_RED|BACKGROUND_BLUE|BACKGROUND_GREEN)) return YC_WHITE;
  if (oc==(FOREGROUND_RED) || 
      oc==(BACKGROUND_RED)) return YC_RED;
  if (oc==(FOREGROUND_BLUE) || 
      oc==(BACKGROUND_BLUE)) return YC_BLUE;
  if (oc==(FOREGROUND_GREEN) || 
      oc==(BACKGROUND_GREEN)) return YC_GREEN;
  if (oc==(FOREGROUND_RED|FOREGROUND_GREEN) || 
      oc==(BACKGROUND_RED|BACKGROUND_GREEN)) return YC_YELLOW;
  if (oc==(FOREGROUND_BLUE|FOREGROUND_GREEN) || 
      oc==(BACKGROUND_BLUE|BACKGROUND_GREEN)) return YC_CYAN;
  if (oc==(FOREGROUND_RED|FOREGROUND_BLUE) || 
      oc==(BACKGROUND_RED|BACKGROUND_BLUE)) return YC_MAGENTA;
#endif
  return -1;
}


void putchars(void) /* put rect onto screen */
{
  char *ch,text,fore[4],back[4];
  int n,sx,sy,x,y,f,b;
  int tox,toy;
  int oldx,oldy;
#ifdef WINDOWS
  CONSOLE_SCREEN_BUFFER_INFO csbi;
  COORD cp;
  char buff[2];
  int written;
#endif

  toy=(int)(pop(stNUMBER)->value);
  tox=(int)(pop(stNUMBER)->value);
  ch=pop(stSTRING)->pointer;
#ifdef UNIX
  getyx(stdscr,oldy,oldx);
#else
  GetConsoleScreenBufferInfo(ConsoleOutput,&csbi);
  oldx=csbi.dwCursorPosition.X;
  oldy=csbi.dwCursorPosition.Y;
#endif
  if (sscanf(ch,"%d,%d:%n",&sx,&sy,&n)!=2) {
    error(ERROR,"illegal screen string");
    return;
  }
  ch+=n;
  for(x=tox;x<tox+sx;x++) {
    if (x<0 || x>=COLS) continue;
    for(y=toy;y<toy+sy;y++) {
      if (x<0 || x>=COLS || y<0 || y>=LINES) {
        for(n=0;n<10;n++) if (*ch) ch++;
        continue;
      }
      if (!*ch) {
        text=' ';
        f=YC_BLACK;
        b=YC_BLACK;
      } else {
        text=*ch;
        strncpy(fore,ch+2,3);
        fore[3]='\0';
        strncpy(back,ch+6,3);
        back[3]='\0';
        for(n=0;n<10;n++) if (*ch) ch++;
        f=name2yc(fore);
        if (f<0) f=YC_WHITE;
        b=name2yc(back);
        if (b<0) b=YC_WHITE;
      }
#ifdef UNIX
      if (has_colors()) attrset(COLOR_PAIR(f*8+b));
      mvaddch(y,x,text);
#else
      cp.X=x;
      cp.Y=y;
      SetConsoleCursorPosition(ConsoleOutput,cp);
      SetConsoleTextAttribute(ConsoleOutput,(WORD)(yc2oc(f,TRUE)|yc2oc(b,FALSE)));
      buff[0]=text;
      buff[1]='\0';
      WriteConsole(ConsoleOutput,buff,1,&written,NULL);
#endif
    }
  }
#ifdef UNIX
    if (has_colors()) 
      attrset(A_NORMAL | COLOR_PAIR(stdfc*8+stdbc));
    else
      attrset(A_NORMAL);
    move(y,x);
    refresh();
#else
    cp.X=oldx;
    cp.Y=oldy;
    SetConsoleCursorPosition(ConsoleOutput,cp);
    SetConsoleTextAttribute(ConsoleOutput,(WORD)(stdfc|stdbc));
#endif

    return;
}


char *getchars(int xf,int yf,int xt,int yt) /* get rect from screen */
{
  int x,y;
#ifdef UNIX
  int c,ct,cc;
#endif
  int cf,cb;
  int oldx,oldy;
  char *res;
  char cols[20];
#ifdef WINDOWS
  CONSOLE_SCREEN_BUFFER_INFO csbi;
  COORD cp;
  char charbuff[2];
  WORD attrbuff[2];
  int read;
#endif

  if (xf>xt) {x=xf;xf=xt;xt=x;}
  if (yf>yt) {y=yf;yf=yt;yt=y;}

  res=my_malloc(12+(xt-xf+1)*(yt-yf+1)*12);
  sprintf(res,"%d,%d:",xt-xf+1,yt-yf+1);
#ifdef UNIX
  getyx(stdscr,oldy,oldx);
#else
  GetConsoleScreenBufferInfo(ConsoleOutput,&csbi);
  oldx=csbi.dwCursorPosition.X;
  oldy=csbi.dwCursorPosition.Y;
#endif

  for(x=xf;x<=xt;x++) {
    for(y=yf;y<=yt;y++) {
      if (y<0 || y>=LINES || x<0 || x>=COLS) {
        strcat(res," blbl");
      } else {
#ifdef UNIX
        c=mvinch(y,x);
        ct=c&A_CHARTEXT;
        if (!isprint(ct)) ct=' ';
        cc=PAIR_NUMBER(c&A_COLOR);
        cb=cc&7;
        cf=(cc-cb)/8;
        if (has_colors()) {
          sprintf(cols,"%c:%s:%s,",ct,yc2short(cf),yc2short(cb));
        } else {
          sprintf(cols,"%c:???:???,",ct);
        }
#else
        cp.X=x;
        cp.Y=y;
        ReadConsoleOutputCharacter(ConsoleOutput,charbuff,1,cp,&read);
        charbuff[1]='\0';
        ReadConsoleOutputAttribute(ConsoleOutput,attrbuff,1,cp,&read);
        cf=oc2yc(attrbuff[0]&(FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_BLUE));
        cb=oc2yc(attrbuff[0]&(BACKGROUND_RED|BACKGROUND_GREEN|BACKGROUND_BLUE));
        sprintf(cols,"%c:%s:%s,",charbuff[0],yc2short(cf),yc2short(cb));
#endif
        strcat(res,cols);
      }
    }
  }
#ifdef UNIX
  move(oldy,oldx);
#endif
  res[strlen(res)-1]='\0';
  return res;
}


char *yc2short(int col) /* convert yabasic colours to short colour name */
{
  static char r1[4],r2[4];
  static char *pr=r1;

  if (pr==r1) 
    pr=r2;
  else
    pr=r1;
  
  strcpy(pr,"***");
  if (col==YC_BLACK) strcpy(pr,"Bla");
  if (col==YC_WHITE) strcpy(pr,"Whi");
  if (col==YC_RED) strcpy(pr,"Red");
  if (col==YC_BLUE) strcpy(pr,"Blu");
  if (col==YC_GREEN) strcpy(pr,"Gre");
  if (col==YC_YELLOW) strcpy(pr,"Yel");
  if (col==YC_CYAN) strcpy(pr,"Cya");
  if (col==YC_MAGENTA) strcpy(pr,"Mag");

  return pr;
}

