%{
/*

    YABASIC ---  a simple Basic Interpreter
    written by Marc-Oliver Ihm 1995-99
    homepage: www.yabasic.de

    BISON part
     
    This file is part of yabasic and may be copied only 
    under the terms of either the Artistic License or 
    the GNU General Public License (GPL), both of which 
    can be found at www.yabasic.de
 
*/

/*
  $Id: yabasic.bison,v 2.55 1999/08/13 16:46:15 ihm Exp ihm $
*/

#undef WINDOWS

#ifndef YABASIC_INCLUDED
#include "yabasic.h"     /* definitions of yabasic */
#endif

#include <malloc.h>

#if HAVE_ALLOCA_H
#include <alloca.h>
#endif

void __yy_bcopy(char *,char *,int); /* prototype missing */

int tileol; /* true, read should go to eon of line */
int yylineno=1; /* line number; counts fresh in every new file */
int main_lineno=1; /* line number of main */
int function_type=ftNONE; /* contains function type while parsing function */
char *current_function=NULL; /* name of currently parsed function */
int exported=FALSE; /* true, if function is exported */
char *due_to_import=NULL; /* library file to import next */
int yylex(void);
extern struct libfile_name *current_libfile; /*  defined in main.c: name of currently parsed file */
%}

%union {
  double fnum;          /* double number */
  int inum;             /* integer number */
  int token;            /* token of command */
  int sep;              /* number of newlines */
  char *string;         /* quoted string */
  char *symbol;         /* general symbol */
  char *digits;         /* string of digits */
  char *docu;		/* embedded documentation */
}

%type <fnum> const
%type <fnum> number
%type <symbol> symbol_or_lineno
%type <symbol> function_name
%type <symbol> function_or_array
%type <symbol> stringfunction_or_array
%type <sep> tSEP
%type <sep> tEOFILE

%token <fnum> tFNUM
%token <symbol> tSYMBOL
%token <symbol> tSTRSYM
%token <symbol> tDOCU
%token <digits> tDIGITS
%token <string> tSTRING

%token tFOR tTO tSTEP tNEXT tWHILE tWEND tREPEAT tUNTIL
%token tGOTO tGOSUB tLABEL tON tSUB tENDSUB tLOCAL tSTATIC tIMPORT tEXPORT tERROR
%token tINTERRUPT tBREAK tCONTINUE tSEP tEOFILE
%token tIF tTHEN tELSE tELSIF tENDIF tUSING
%token tPRINT tINPUT tLINE tRETURN tDIM tEND tEXIT tAT tSCREEN 
%token tREVERSE tCOLOUR
%token tAND tOR tNOT
%token tNEQ tLEQ tGEQ tLTN tGTN tEQU tPOW
%token tREAD tDATA tRESTORE
%token tOPEN tCLOSE tORIGIN
%token tWINDOW tDOT tLINE tCIRCLE tTEXT tCLEAR tPRINTER
%token tWAIT tBELL tARDIM tARSIZE
%token tBOX tRECT tGETBIT tPUTBIT tGETCHAR tPUTCHAR tFRAME tNEW tCURVE

%token tSIN tASIN tCOS tACOS tTAN tATAN tEXP tLOG 
%token tSQRT tSQR tMYEOF tABS tSIG
%token tINT tFRAC tMOD tRAN tLEN tVAL tLEFT tRIGHT tMID tLEN tMIN tMAX
%token tSTR tINKEY tCHR tASC tHEX tDEC tUPPER tLOWER tMOUSEX tMOUSEY tMOUSEB
%token tTRIM tLTRIM tRTRIM tINSTR 
%token tSYSTEM tSYSTEM2 tPEEK tPEEK2 tPOKE 
%token tDATE tTIME tTOKEN tSPLIT tGLOB

%left tOR
%left tAND
%left tNOT

%left '-' '+'
%left '*' '/'
%left tPOW
%nonassoc UMINUS

%%

program: statement_list tEOFILE {end_of_file=TRUE;YYACCEPT;}
  ;

statement_list: statement
  | statement_list {if (errorlevel<=ERROR) {YYABORT;}} 
    tSEP {yylineno+=$3;
	  if (due_to_import) {import_lib(due_to_import);due_to_import=NULL;}
	} statement
  ;

statement:  /* empty */
  | string_assignment 
  | assignment
  | tIMPORT tSYMBOL {due_to_import=my_strdup($2);}
  | tERROR string_expression {add_command(cERROR,NULL);}
  | for_loop 
  | repeat_loop
  | while_loop
  | function_definition
  | function_or_array {create_call($1);add_command(cPOP,NULL);}
  | stringfunction_or_array {create_call($1);add_command(cPOPSTRSYM,NULL);}
  | tLOCAL {if (function_type==ftNONE) error(ERROR,"no use for 'local' outside functions");} local_list
  | tSTATIC {if (function_type==ftNONE) error(ERROR,"no use for 'static' outside functions");} static_list
  | if_clause
  | short_if
  | tGOTO symbol_or_lineno {create_goto($2);}
  | tGOSUB symbol_or_lineno {create_gosub($2);}
  | tON tINTERRUPT tBREAK {create_exception(TRUE);}
  | tON tINTERRUPT tCONTINUE {create_exception(FALSE);}
  | tON expression tGOTO {add_command(cSKIPPER,NULL);}
    goto_list {add_command(cNOP,NULL);}
  | tON expression tGOSUB {add_command(cSKIPPER,NULL);} 
    gosub_list {add_command(cNOP,NULL);}
  | tLABEL symbol_or_lineno {create_label($2,cLABEL);}
  | open_clause {add_command(cCHECKOPEN,NULL);}
  | tCLOSE hashed_number {add_command(cCLOSE,NULL);}
  | tPRINT printintro printlist {create_colour(0);create_print('n');} 
  | tPRINT printintro printlist ';' {create_colour(0);}
  | tPRINT printintro printlist ',' {create_colour(0);create_print('t');}
  | tINPUT {tileol=FALSE;} inputintro
  | tLINE tINPUT {tileol=TRUE;} inputintro
  | tREAD readlist
  | tDATA datalist
  | tRESTORE {create_restore("");}
  | tRESTORE symbol_or_lineno {create_restore($2);}
  | tRETURN {if (function_type!=ftNONE) {
	       add_command(cPOPSYMLIST,NULL);
               create_retval(ftNONE,function_type);
               add_command(cRET_FROM_FUN,NULL);
            } else {
               add_command(cRETURN,NULL);
            }}
  | tRETURN expression {if (function_type==ftNONE) {error(ERROR,"can not return value"); YYABORT;} add_command(cPOPSYMLIST,NULL);create_retval(ftNUMBER,function_type);add_command(cRET_FROM_FUN,NULL);}
  | tRETURN string_expression {if (function_type==ftNONE) {error(ERROR,"can not return value"); YYABORT;} add_command(cPOPSYMLIST,NULL);create_retval(ftSTRING,function_type);add_command(cRET_FROM_FUN,NULL);}
  | tDIM dimlist
  | tOPEN tWINDOW expression ',' expression {create_openwin(FALSE);}
  | tOPEN tWINDOW expression ',' expression ',' string_expression 
         {create_openwin(TRUE);}
  | tWINDOW tORIGIN string_expression {add_command(cMOVEORIGIN,NULL);}
  | tDOT coordinates {add_command(cDOT,NULL);}
  | tLINE coordinates to coordinates {create_line('l');}
  | tLINE to coordinates {create_line('t');}
  | tPUTBIT string_expression to expression ',' expression ',' string_expression {add_command(cPUTBIT,NULL);}
  | tPUTBIT string_expression to expression ',' expression {create_pushstr("replace"); add_command(cPUTBIT,NULL);}
  | tPUTCHAR string_expression to expression ',' expression {add_command(cPUTCHAR,NULL);}
  | tNEW tCURVE {create_line('n');}
  | tCIRCLE coordinates ',' expression {add_command(cCIRCLE,NULL)}
  | tTEXT coordinates ',' string_expression {create_text(FALSE);}
  | tTEXT coordinates ',' string_expression ',' string_expression
        {create_text(TRUE);}
  | tBOX coordinates to coordinates {create_rect(3);}
  | tFRAME coordinates to coordinates {create_rect(2);}
  | tRECT coordinates to coordinates {create_rect(2);}
  | tCLEAR tRECT coordinates to coordinates {create_rect(1);}
  | tCLEAR tBOX coordinates to coordinates {create_rect(1);}
  | tCLOSE tWINDOW {add_command(cCLOSEWIN,NULL);}
  | tCLEAR tWINDOW {add_command(cCLEARWIN,NULL);}
  | tCLEAR tSCREEN {add_command(cCLEARSCR,NULL);}
  | tOPEN tPRINTER {create_openprinter(0);}
  | tOPEN tPRINTER string_expression {create_openprinter(1);}
  | tCLOSE tPRINTER {add_command(cCLOSEPRN,NULL);}
  | tWAIT expression {add_command(cWAIT,NULL);}
  | tBELL {add_command(cBELL,NULL);}
  | tINKEY {create_pushdbl(-1);create_function(fINKEY);add_command(cPOPSTRSYM,NULL);}
  | tINKEY '(' ')' {create_pushdbl(-1);create_function(fINKEY);add_command(cPOPSTRSYM,NULL);}
  | tINKEY '(' expression ')' {create_function(fINKEY);add_command(cPOPSTRSYM,NULL);}
  | tSYSTEM2 '(' string_expression ')' {create_function(fSYSTEM2);
	add_command(cPOP,NULL);}
  | tPOKE string_expression ',' string_expression {create_poke('s');}
  | tPOKE string_expression ',' expression {create_poke('d');}
  | tPOKE hashed_number ',' string_expression {create_poke('S');}
  | tPOKE hashed_number ',' expression {create_poke('D');}
  | tEND {add_command(cEND,NULL);}
  | tEXIT {create_pushdbl(0.);add_command(cEXIT,NULL);}
  | tEXIT expression {add_command(cEXIT,NULL);}
  | tDOCU {create_docu($1);}
  ;

string_assignment: tSTRSYM tEQU string_expression {add_command(cPOPSTRSYM,dotify($1));}
  | tMID '(' string_scalar_or_array ',' expression ',' expression ')' tEQU string_expression {create_changestring(fMID);}
  | tLEFT '(' string_scalar_or_array ',' expression ')' tEQU string_expression {create_changestring(fLEFT);}
  | tRIGHT '(' string_scalar_or_array ',' expression ')' tEQU string_expression {create_changestring(fRIGHT);}
  | stringfunction_or_array tEQU string_expression {create_doarray(dotify($1),ASSIGNSTRINGARRAY);}
  ;

to: ','
  | tTO
  ;


open_clause: tOPEN hashed_number ',' string_expression ',' string_expression
    {create_myopen(3);}
  | tOPEN hashed_number ',' string_expression {create_myopen(2);}
  ;


string_scalar_or_array: tSTRSYM {add_command(cPUSHSTRPTR,dotify($1));}
  | tSTRSYM '(' call_list ')' {create_doarray(dotify($1),GETSTRINGPOINTER);} 
  ;

string_expression: tSTRSYM {add_command(cPUSHSTRSYM,dotify($1));}
  | string_function
  | stringfunction_or_array {create_function_or_array(cSTRINGFUNCTION_OR_ARRAY,$1);}
  | tSTRING {if ($1==NULL) {error(ERROR,"String not terminated");create_pushstr("");} else {create_pushstr($1);}}
  | string_expression '+' string_expression {add_command(cCONCAT,NULL);}
  | '(' string_expression ')'
  ;

string_function: tLEFT '(' string_expression ',' expression ')' {create_function(fLEFT);}
  | tRIGHT '(' string_expression ',' expression ')' {create_function(fRIGHT);}
  | tMID '(' string_expression ',' expression ',' expression ')' {create_function(fMID);}
  | tSTR '(' expression ')' {create_function(fSTR);}
  | tSTR '(' expression ',' string_expression ')' {create_function(fSTR2);} 
  | tINKEY {create_pushdbl(-1);create_function(fINKEY);}
  | tINKEY '(' ')' {create_pushdbl(-1);create_function(fINKEY);}
  | tINKEY '(' expression ')' {create_function(fINKEY);}
  | tCHR '(' expression ')' {create_function(fCHR);}
  | tUPPER '(' string_expression ')' {create_function(fUPPER);}
  | tLOWER '(' string_expression ')' {create_function(fLOWER);}
  | tLTRIM '(' string_expression ')' {create_function(fLTRIM);}
  | tRTRIM '(' string_expression ')' {create_function(fRTRIM);}
  | tTRIM '(' string_expression ')' {create_function(fTRIM);}
  | tSYSTEM '(' string_expression ')' {create_function(fSYSTEM);}
  | tDATE {create_function(fDATE);}
  | tTIME {create_function(fTIME);}
  | tPEEK2 '(' string_expression ')' {create_function(fPEEK2);}
  | tPEEK2 '(' string_expression ',' string_expression ')' {create_function(fPEEK3);}
  | tTOKEN '(' string_scalar_or_array ',' string_expression ')' {create_token(2);}
  | tTOKEN '(' string_scalar_or_array ')' {create_token(0);}
  | tSPLIT '(' string_scalar_or_array ',' string_expression ')' {create_token(3);}
  | tSPLIT '(' string_scalar_or_array ')' {create_token(1);}
  | tGETBIT '(' coordinates to coordinates ')' {create_function(fGETBIT);}
  | tGETCHAR '(' expression ',' expression to expression ',' expression ')' {create_function(fGETCHAR);}
  | tHEX '(' expression ')' {create_function(fHEX);}
  ;

assignment: tSYMBOL tEQU expression {add_command(cPOPDBLSYM,dotify($1));} 
  | function_or_array tEQU expression {create_doarray($1,ASSIGNARRAY);}
  ;

expression: number {create_pushdbl($1);}
  | function
  | tARDIM '(' expression ')' {add_command(cARDIM,NULL);}
  | tARDIM '(' string_expression ')' {add_command(cARDIM,NULL);}
  | tARSIZE '(' expression ',' expression ')' {add_command(cARSIZE,NULL);}
  | tARSIZE '(' string_expression ',' expression ')' {add_command(cARSIZE,NULL);}
  | function_or_array {create_function_or_array(cFUNCTION_OR_ARRAY,$1);}
  | '(' expression ')'
  | tSYMBOL {add_command(cPUSHDBLSYM,dotify($1));}
  | expression '+' expression {create_dblbin('+');}
  | expression '-' expression {create_dblbin('-');}
  | expression '*' expression {create_dblbin('*');}
  | expression '/' expression {create_dblbin('/');}
  | expression tPOW expression {create_dblbin('^');}
  | '-' expression %prec UMINUS {add_command(cNEGATE,NULL);}
  ;

coordinates: expression ',' expression
  ;

function: tSIN '(' expression ')' {create_function(fSIN);}
  | tASIN '(' expression ')' {create_function(fASIN);}
  | tCOS '(' expression ')' {create_function(fCOS);}
  | tACOS '(' expression ')' {create_function(fACOS);}
  | tTAN '(' expression ')' {create_function(fTAN);}
  | tATAN '(' expression ')' {create_function(fATAN);}
  | tATAN '(' expression ',' expression  ')' {create_function(fATAN2);}
  | tEXP '(' expression ')' {create_function(fEXP);}
  | tLOG '(' expression ')' {create_function(fLOG);}
  | tSQRT '(' expression ')' {create_function(fSQRT);}
  | tSQR '(' expression ')' {create_function(fSQR);}
  | tINT '(' expression ')' {create_function(fINT);}
  | tFRAC '(' expression ')' {create_function(fFRAC);}
  | tABS '(' expression ')' {create_function(fABS);}
  | tSIG '(' expression ')' {create_function(fSIG);}
  | tMOD '(' expression ',' expression ')' {create_function(fMOD);}
  | tRAN '(' expression ')' {create_function(fRAN);}
  | tRAN '(' ')' {create_function(fRAN2);}
  | tMIN '(' expression ',' expression ')' {create_function(fMIN);}
  | tMAX '(' expression ',' expression ')' {create_function(fMAX);}
  | tLEN '(' string_expression ')' {create_function(fLEN);}
  | tVAL '(' string_expression ')' {create_function(fVAL);}
  | tASC '(' string_expression ')' {create_function(fASC);}
  | tDEC '(' string_expression ')' {create_function(fDEC);}
  | tINSTR '(' string_expression ',' string_expression ')' {create_function(fINSTR);}
  | tSYSTEM2 '(' string_expression ')' {create_function(fSYSTEM2);}
  | tPEEK '(' hashed_number ')' {create_function(fPEEK4);}
  | tPEEK '(' string_expression ')' {create_function(fPEEK);}
  | tMOUSEX '(' string_expression ')' {create_function(fMOUSEX);}
  | tMOUSEX {create_pushstr("");create_function(fMOUSEX);}
  | tMOUSEY '(' string_expression ')' {create_function(fMOUSEY);}
  | tMOUSEY {create_pushstr("");create_function(fMOUSEY);}
  | tMOUSEB '(' string_expression ')' {create_function(fMOUSEB);}
  | tMOUSEB {create_pushstr("");create_function(fMOUSEB);}
  | tOPEN '(' string_expression ')' {create_myopen(-1);}
  | tOPEN '(' string_expression ',' string_expression ')' {create_myopen(-2);}
  ;

const: number {$$=$1;}
  | '+' number {$$=$2;}
  | '-' number {$$=-$2;}
  ;

number: tFNUM {$$=$1;}
  | tDIGITS {$$=atoi($1);}
  ;

symbol_or_lineno: tDIGITS {$$=my_strdup(dotify($1));}
  | tSYMBOL {$$=my_strdup(dotify($1));}
  ;

dimlist: tSYMBOL '(' call_list ')' {create_dim(dotify($1),'d');}
  | dimlist ',' tSYMBOL '(' call_list ')' {create_dim(dotify($3),'d');}
  | tSTRSYM '(' call_list ')' {create_dim(dotify($1),'s');}
  | dimlist ',' tSTRSYM '(' call_list ')' {create_dim(dotify($3),'s');}
  ;

function_or_array: tSYMBOL '(' call_list ')' {$$=my_strdup(dotify($1));}
  ;

stringfunction_or_array: tSTRSYM '(' call_list ')' {$$=my_strdup(dotify($1));}
  ;

call_list: {add_command(cPUSHFREE,NULL);} calls
  ;

calls: /* empty */
  | call_item
  | calls ',' call_item
  ;

call_item: string_expression
  | expression
  ;
 
function_definition: export tSUB {pushlabel();if (function_type!=ftNONE) {error(ERROR,"nested functions not allowed");YYABORT;}}
	function_name {if (exported) create_sublink($4); create_label($4,cUSER_FUNCTION);
	               add_command(cPUSHSYMLIST,NULL);add_command(cNUMPARAM,NULL);}
	'(' paramlist ')' {create_require(stFREE);add_command(cPOP,NULL);}
	statement_list
	tENDSUB {add_command(cPOPSYMLIST,NULL);create_retval(ftNONE,function_type);function_type=ftNONE;add_command(cRET_FROM_FUN,NULL);create_endfunction();poplabel();}

function_name: tSYMBOL {function_type=ftNUMBER;current_function=my_strdup(dotify($1));$$=my_strdup(dotify($1));}
  | tSTRSYM {function_type=ftSTRING;current_function=my_strdup(dotify($1));$$=my_strdup(dotify($1));}
  ;

export: /* empty */ {exported=FALSE;}
  | tEXPORT {exported=TRUE;}

local_list: local_item
  | local_list ',' local_item
  ;

local_item: tSYMBOL {create_makelocal(dotify($1),syNUMBER);}
  | tSTRSYM {create_makelocal(dotify($1),sySTRING);}
  | tSYMBOL '(' call_list ')' {create_makelocal(dotify($1),syARRAY);create_dim(dotify($1),'d');}
  | tSTRSYM '(' call_list ')' {create_makelocal(dotify($1),syARRAY);create_dim(dotify($1),'s');}
  ;
  
static_list: static_item
  | static_list ',' static_item
  ;

static_item: tSYMBOL {create_makestatic(dotify($1),syNUMBER);}
  | tSTRSYM {create_makestatic(dotify($1),sySTRING);}
  | tSYMBOL '(' call_list ')' {create_makestatic(dotify($1),syARRAY);create_dim(dotify($1),'d');}
  | tSTRSYM '(' call_list ')' {create_makestatic(dotify($1),syARRAY);create_dim(dotify($1),'s');}
  ;
  
paramlist: /* empty */
  | paramitem
  | paramlist ',' paramitem
  ;
  
paramitem: tSYMBOL {create_require(stNUMBER);create_makelocal(dotify($1),syNUMBER);add_command(cPOPDBLSYM,dotify($1));}
  | tSTRSYM {create_require(stSTRING);create_makelocal(dotify($1),sySTRING);add_command(cPOPSTRSYM,dotify($1));}
  | tSYMBOL '(' ')' {create_require(stNUMBERARRAYREF);create_arraylink(dotify($1),stNUMBERARRAYREF);}
  | tSTRSYM '(' ')' {create_require(stSTRINGARRAYREF);create_arraylink(dotify($1),stSTRINGARRAYREF);}
  ;

for_loop: tFOR tSYMBOL tEQU 
            {pushname(dotify($2)); /* will be used by next_symbol to check equality */
	     add_command(cRESETSKIPONCE,NULL);
	     pushgoto();}
	  expression tTO expression 
	  step_part { /* pushes another expression */
	     add_command(cSKIPONCE,NULL);
	     pushlabel();
	     add_command(cSTARTFOR,NULL);
	     add_command(cPOPDBLSYM,dotify($2));
	     poplabel();
	     add_command(cPUSHDBLSYM,dotify($2));
	     add_command(cFORINCREMENT,NULL);
	     add_command(cPOPDBLSYM,dotify($2));
	     add_command(cPUSHDBLSYM,dotify($2));
	     add_command(cFORCHECK,NULL);
	     add_command(cDECIDE,NULL);
             pushlabel();}
	  tSEP {yylineno+=$10;}
          statement_list {
             swap();popgoto();poplabel();}
          next_or_eofile next_symbol
  ;

next_or_eofile: tNEXT
  | tEOFILE {end_of_file=TRUE;
	    error(ERROR,"'next' statement is missing"); YYABORT;}

step_part: {create_pushdbl(1.0);} /* can be omitted */
  | tSTEP expression
  ;

next_symbol:  {pop(stSTRING);}/* can be omitted */
  | tSYMBOL {if (strcmp(pop(stSTRING)->pointer,dotify($1))) 
             {error(ERROR,"'for' and 'next' do not match"); YYABORT;}
           }
  ;

while_loop: tWHILE {pushgoto()} '(' condition ')'
	      {add_command(cDECIDE,NULL);
	      pushlabel();}
	      statement_list
	    tWEND
	      {swap();popgoto();poplabel();}
  ;	    

repeat_loop: tREPEAT {pushgoto();} 
	       statement_list 
             tUNTIL '(' condition ')'
	       {add_command(cDECIDE,NULL);popgoto();}
  ;

if_clause: tIF condition {add_command(cDECIDE,NULL);storelabel();pushlabel();}
           tTHEN statement_list {swap();matchgoto();swap();poplabel();}
	   elsif_part
           else_part {poplabel();}
           endif_or_eof
  ;

short_if:  tIF condition {add_command(cDECIDE,NULL);pushlabel();}
	statement {poplabel();}
  ;

endif_or_eof: tENDIF
  | tEOFILE {end_of_file=TRUE;
            error(ERROR,"'endif'-statement is missing"); YYABORT;}

condition: '(' condition ')'
  | condition tOR condition {create_boole('|');}
  | condition tAND condition {create_boole('&');}
  | tNOT condition {create_boole('!');}
  | string_expression tEQU string_expression {create_strrelop('=');}
  | string_expression tNEQ string_expression {create_strrelop('!');}
  | string_expression tLTN string_expression {create_strrelop('<');}
  | string_expression tLEQ string_expression {create_strrelop('{');}
  | string_expression tGTN string_expression {create_strrelop('>');}
  | string_expression tGEQ string_expression {create_strrelop('}');}
  | expression tEQU expression {create_dblrelop('=');}
  | expression tNEQ expression {create_dblrelop('!');}
  | expression tLTN expression {create_dblrelop('<');}
  | expression tLEQ expression {create_dblrelop('{');}
  | expression tGTN expression {create_dblrelop('>');}
  | expression tGEQ expression {create_dblrelop('}');}
  | tMYEOF '(' expression ')' {add_command(cTESTEOF,NULL);}
  | tGLOB '(' string_expression ',' string_expression ')' {add_command(cGLOB,NULL);}
  | open_clause 
  ;

else_part: /* can be omitted */
  | tELSE statement_list
  ;

elsif_part: /* can be omitted */
  | tELSIF condition tTHEN
    	{add_command(cDECIDE,NULL);pushlabel();} 
    statement_list 
	{swap();matchgoto();swap();poplabel();} 
    elsif_part
  ;

inputlist: input
  | input ',' {add_command(cCHKPROMPT,NULL);} inputlist
  ;

input: tSYMBOL {create_myread('d',tileol);add_command(cPOPDBLSYM,dotify($1));}
  | tSYMBOL '(' call_list ')' 
    	{create_myread('d',tileol);create_doarray(dotify($1),ASSIGNARRAY);}
  | tSTRSYM {create_myread('s',tileol);add_command(cPOPSTRSYM,dotify($1));}
  | tSTRSYM '(' call_list ')' 
    	{create_myread('s',tileol);create_doarray(dotify($1),ASSIGNSTRINGARRAY);}
  ;

readlist: readitem
  | readlist ',' readitem
  ;

readitem: tSYMBOL {create_readdata('d');add_command(cPOPDBLSYM,dotify($1));}
  | tSYMBOL '(' call_list ')' 
    {create_readdata('d');create_doarray(dotify($1),ASSIGNARRAY);}
  | tSTRSYM {create_readdata('s');add_command(cPOPSTRSYM,dotify($1));}
  | tSTRSYM '(' call_list ')' 
    {create_readdata('s');create_doarray(dotify($1),ASSIGNSTRINGARRAY);}
  ;

datalist: tSTRING {create_strdata($1);}
  | const {create_dbldata($1);}
  | datalist ','  tSTRING {create_strdata($3);}
  | datalist ',' const {create_dbldata($3);}
  ;

printlist:  /* possible empty */
  | expression using
  | printlist ',' expression using
  | string_expression {create_print('s');} 
  | printlist ',' string_expression {create_print('s');}
  ;

using: {create_print('d');} /* possible empty */
  | tUSING string_expression {create_print('u');}
  ;

inputintro: stream {create_myswitch(-1);} inputlist
  | position {create_myswitch(0);} prompt inputlist
  ;

prompt: /* empty */ {create_pushstr("?");create_print('s');} 
  | tSTRING {create_pushstr($1);create_print('s');}
  ;

printintro: stream {create_myswitch(-1);}
  | position {create_myswitch(0);}
  | tREVERSE position {create_colour(1);create_myswitch(0);}
  | tCOLOUR '(' string_expression ')' position {create_colour(2);create_myswitch(0);}
  | tCOLOUR '(' string_expression ',' string_expression ')' position {create_colour(3);create_myswitch(0);}
  ;  

position: /* can be empty */
  | tAT '(' expression ',' expression ')' {add_command(cMOVE,NULL);}
  ;

stream: '#' '(' expression ')'
  | '#' intnum
  ;

hashed_number: '#' expression
  | expression
  ;

intnum: tDIGITS {create_pushdbl(atoi($1));}

goto_list: symbol_or_lineno {create_goto($1);add_command(cFINDNOP,NULL);}
  | goto_list ',' symbol_or_lineno {create_goto($3);add_command(cFINDNOP,NULL);}
  ;

gosub_list: symbol_or_lineno {create_gosub($1);add_command(cFINDNOP,NULL);}
  | gosub_list ',' symbol_or_lineno {create_gosub($3);add_command(cFINDNOP,NULL);}
  ;
