%{
/*

    YABASIC ---  a simple Basic Interpreter
    written by Marc-Oliver Ihm 1995-99
    homepage: www.yabasic.de

    FLEX part
     
    This file is part of yabasic and may be copied only 
    under the terms of either the Artistic License or 
    the GNU General Public License (GPL), both of which 
    can be found at www.yabasic.de

*/

/*
  $Id: yabasic.flex,v 2.55 1999/08/13 16:46:22 ihm Exp ihm $
*/

#include <string.h>
#undef WINDOWS
#undef yywrap

#include "bison.h"       /* get tokens from BISON */
#ifndef YABASIC_INCLUDED
#include "yabasic.h"     /* definitions of yabasic */
#endif
extern int main_lineno; /* defined in yabasic.bison: line number of main file */
extern int yylineno; /* defined in yabasic.bison: line number of main file */
extern struct libfile_name *libfile_root; /* defined in main.c: first file, i.e. name of main file */

YY_BUFFER_STATE main_buffer=NULL; /* Lex buffer for main file */
YY_BUFFER_STATE lib_buffer=NULL; /* Lex buffer for any imported file */
%}

WS [ \t\f\r\v]
NAME ([A-Za-z_][A-za-z0-9_]*\.[A-Za-z_][A-za-z0-9_]*)|([A-Za-z_][A-za-z0-9_]*)

%option noyywrap
%x RET
%x PRELNO
%x PASTLNO
%s PASTNEWLINE

%%
{WS}* {BEGIN INITIAL;}     /* ignore whitespace */
<RET>{WS}* {BEGIN INITIAL;} /* still ignore whitespace */

\\{WS}*\n {yylineno++;BEGIN PASTNEWLINE;} /* allow for escaped newlines */
<RET>\\{WS}*\n {yylineno++;BEGIN PASTNEWLINE;}

<INITIAL>^{WS}*[0-9]* {BEGIN PRELNO;yyless(0);return tLABEL;}
<RET>^{WS}*[0-9]* {BEGIN PRELNO;yyless(0);return tLABEL;}
<PRELNO>{WS}* {}
<PRELNO>[0-9]* {
  BEGIN PASTLNO;
  yylval.symbol=(char *)my_strdup(yytext);
  return tSYMBOL;
}
<PASTLNO>{WS} {BEGIN INITIAL; yylval.sep=0;return tSEP;}
<PASTLNO>\n {BEGIN INITIAL;yylval.sep=1;return tSEP;}
<PRELNO>\n {BEGIN INITIAL;yylval.sep=1;return tSEP;}
<PASTLNO>. {return yytext[0];}
<PRELNO>. {return yytext[0];}

<<EOF>> {yylval.sep=0;return tEOFILE;}
<RET>\n {yylval.sep=1;return (interactive && !in_libfile) ? tEOFILE:tSEP;}
<RET>[^\n] {BEGIN INITIAL;unput(yytext[0]);}
\n {BEGIN RET;yylval.sep=1;return tSEP;}
: {yylval.sep=0;return tSEP;}

REM{WS}+.*\n {yylval.sep=1;return tSEP;}  /* comments span 'til end of line */
REM\n {yylval.sep=1;return tSEP;}  /* comments span 'til end of line */

((DOCU|DOC|DOCUMENTATION)({WS}+.*)?) {
  yylval.docu=(char *)my_strdup(strpbrk(yytext," \t\r\f\v")+1);
  return tDOCU;
}

^#.*\n {yylval.sep=1;return tSEP;} /* '#' as first character may introduce comments too */

END{WS}+SUB return tENDSUB;
END{WS}+IF return tENDIF;
END{WS}+WHILE return tWEND;
IMPORT return tIMPORT;
EXPORT return tEXPORT;
ERROR return tERROR;
FOR return tFOR;
TO return tTO;
STEP return tSTEP;
NEXT return tNEXT;
WHILE return tWHILE;
WEND return tWEND;
REPEAT return tREPEAT;
UNTIL return tUNTIL;
GOTO return tGOTO;
GOSUB return tGOSUB;
SUB return tSUB;
SUBROUTINE return tSUB;
LOCAL return tLOCAL;
STATIC return tSTATIC;
ON return tON;
INTERRUPT return tINTERRUPT;
BREAK return tBREAK;
CONTINUE return tCONTINUE;
LABEL return tLABEL;
IF return tIF;
THEN return tTHEN;
ELSE return tELSE;
ELSIF return tELSIF;
ELSEIF return tELSIF;
ENDIF return tENDIF;
FI return tENDIF;
OPEN return tOPEN;
CLOSE return tCLOSE;
PRINT return tPRINT;
USING return tUSING;
REVERSE return tREVERSE;
COLOR return tCOLOUR;
COLOUR return tCOLOUR;
\? return tPRINT;
INPUT return tINPUT;
RETURN return tRETURN;
DIM return tDIM;
REDIM return tDIM;
END return tEND;
EXIT return tEXIT;
READ return tREAD;
DATA return tDATA;
RESTORE return tRESTORE;
AND return tAND;
OR return tOR;
NOT return tNOT;
WINDOW return tWINDOW;
ORIGIN return tORIGIN;
PRINTER return tPRINTER;
DOT return tDOT;
LINE return tLINE;
CURVE return tCURVE;
CIRCLE return tCIRCLE;
CLEAR return tCLEAR;
TEXT return tTEXT;
BOX return tBOX;
FRAME return tFRAME;
RECTANGLE return tRECT;
RECT return tRECT;
BITBLIT return tPUTBIT;
BITBLT return tPUTBIT;
PUTBIT return tPUTBIT;
"BITBLT$" return tGETBIT;
"BITBLIT$" return tGETBIT;
"GETBIT$" return tGETBIT;
PUTSCREEN return tPUTCHAR;
"GETSCREEN$" return tGETCHAR;
NEW return tNEW;
WAIT return tWAIT;
PAUSE return tWAIT;
BELL return tBELL;
BEEP return tBELL;
ARRAYDIM return tARDIM;
ARRAYDIMENSION return tARDIM;
ARRAYSIZE return tARSIZE;
NUMPARAM(S)?({WS}*\({WS}*\))? {yylval.symbol=(char *)my_strdup("numparams"); return tSYMBOL;}

SIN return tSIN;
ASIN return tASIN;
COS return tCOS;
ACOS return tACOS;
TAN return tTAN;
ATAN return tATAN;
EXP return tEXP;
LOG return tLOG;
SQRT return tSQRT;
SQR return tSQR;
INT return tINT;
FRAC return tFRAC;
ABS return tABS;
SIG return tSIG;
MOD return tMOD;
RAN return tRAN;
MIN return tMIN;
MAX return tMAX;
"LEFT$" return tLEFT;
"RIGHT$" return tRIGHT;
"MID$" return tMID;
"LOWER$" return tLOWER;
"UPPER$" return tUPPER;
"LTRIM$" return tLTRIM;
"RTRIM$" return tRTRIM;
"TRIM$" return tTRIM;
INSTR return tINSTR;
LEN return tLEN;
VAL return tVAL;
EOF return tMYEOF;
"STR$" return tSTR;
"INKEY$" return tINKEY;
"MOUSEX" return tMOUSEX;
"MOUSEY" return tMOUSEY;
"MOUSEB" return tMOUSEB;
"MOUSEBUTTON" return tMOUSEB;
"CHR$" return tCHR;
ASC return tASC;
"HEX$" return tHEX;
DEC return tDEC;
AT return tAT;
@ return tAT;
SCREEN return tSCREEN;
"SYSTEM$" return tSYSTEM;
SYSTEM return tSYSTEM2;
"DATE$" return tDATE;
"TIME$" return tTIME;
PEEK return tPEEK;
"PEEK$" return tPEEK2;
POKE return tPOKE;
"TOKEN$" return tTOKEN;
"SPLIT$" return tSPLIT;
GLOB return tGLOB;
"^" return tPOW;
"**" return tPOW;

"<>" return tNEQ;
"<=" return tLEQ;
">=" return tGEQ;
"=" return tEQU;
"<" return tLTN;
">" return tGTN;
"!" return tNOT;

[-+*/:(),.;] {return yytext[0];}

[0-9]+ {
  yylval.digits=(char *)my_strdup(yytext);
  return tDIGITS;
}

(([0-9]+|([0-9]*\.[0-9]*))([eE][-+]?[0-9]+)?) {
  { double d;
    sscanf(yytext,"%lg",&d);
    yylval.fnum=d;
    return tFNUM;
  }
}

{NAME} {
  yylval.symbol=(char *)my_strdup(yytext);
  return tSYMBOL;
}

  /* Symbols with a trailing $-sign are treated special */
{NAME}\$ {
  yylval.symbol=(char *)my_strdup(yytext);
  return tSTRSYM;
}

\"[^"]*(\"|\n) {
  int cnt;
  if (yyleng[yytext-1]=='\n') {
  	yylval.string=NULL;
  	return tSTRING;
  }
  for(cnt=0;yytext[yyleng-cnt-2]=='\\';cnt++) ;
  if (cnt%2) {
  	yyless(yyleng-1);
	yymore();
  } else {
	yylval.string=(char *)my_strdup(yytext+1);
	*(yylval.string+yyleng-2)='\0';
	replace(yylval.string);
	return tSTRING;
  }
}


. {if (isprint(yytext[0])) return yytext[0]; else return ' ';}

%%
void yyerror(char *msg)
{
  int i,j;
  char c=' ';	
  
  sprintf(string,"%s at %n",msg,&j);
  if (*yytext=='\n') {
    sprintf(string+j,"end of line");
  }
  else {
    i=0;
    string[j++]='\"';
    while(yytext[i]) {
      if (isprint(yytext[i])) string[j++]=yytext[i++];
      else {
	sprintf(string+j,"0x%02x",yytext[i]);
	j+=4;
	break;
      }
    }
    string[j++]='\"';
    string[j]='\0';
  }
  error(ERROR,string);

  if (c!=' ') unput((int)c); /* stop gcc from complaining about yyunput ... */

  return;
}


int open_main(FILE *file) /* open main file */
{
  main_buffer=yy_create_buffer(file,YY_BUF_SIZE);
  yy_switch_to_buffer(main_buffer);

  return TRUE;
}


int import_lib(char *name) /* import library */
{
  char *full;
  static FILE *file=NULL;

  if (lib_buffer) {
    error(ERROR,"nested imports not allowed");
    return FALSE;
  }
  if (strchr(name,'.')) {
    sprintf(string,"library name '%s' contains '.'",name);
    error(ERROR,string);
    return FALSE;
  }
  if (!strcmp(name,"main")) {
    error(ERROR,"invalid library name 'main'");
    return FALSE;
  }
  
  if (file) {
    fclose(file);
    file=NULL;
  }

  file=open_library(name,&full,FALSE);
  if (!file) return FALSE;
  lib_buffer=yy_create_buffer(file,YY_BUF_SIZE);
  yy_switch_to_buffer(lib_buffer);
  if (infolevel>=NOTE) {
    sprintf(string,"importing file '%s'",full);
    error(NOTE,string);
  }
  current_libfile=new_file(full,NULL);
  if (!current_libfile) {
    sprintf(string,"library '%s' has already been imported",full);
    error(ERROR,string);
    return FALSE;
  } 

  /* start line numbers anew */
  main_lineno=yylineno;
  yylineno=1;
  in_libfile=TRUE;

  return TRUE;
}


int yywrap(void) /* switch back to main_buffer or report end of file */
{
  if (lib_buffer) {
    yy_delete_buffer(lib_buffer);
    lib_buffer=NULL;
    yy_switch_to_buffer(main_buffer);
    current_libfile=libfile_root; /* switch back to main file */
    yylineno=main_lineno;
    in_libfile=FALSE;
    return 0;
  }
  return 1;
}


